// Geometric Tools LLC, Redmond WA 98052
// Copyright (c) 1998-2014
// Distributed under the Boost Software License, Version 1.0.
// http://www.boost.org/LICENSE_1_0.txt
// http://www.geometrictools.com/License/Boost/LICENSE_1_0.txt
// File Version: 1.7.0 (2014/12/08)

//----------------------------------------------------------------------------
template <int N, typename Real>
OrientedBox<N, Real>::OrientedBox()
{
    center.MakeZero();
    for (int i = 0; i < N; ++i)
    {
        axis[i].MakeUnit(i);
        extent[i] = (Real)1;
    }
}
//----------------------------------------------------------------------------
template <int N, typename Real>
OrientedBox<N, Real>::OrientedBox(Vector<N, Real> const& inCenter,
    std::array<Vector<N, Real>, N> const& inAxis,
    Vector<N, Real> const& inExtent)
    :
    center(inCenter),
    axis(inAxis),
    extent(inExtent)
{
}
//----------------------------------------------------------------------------
template <int N, typename Real>
void OrientedBox<N, Real>::GetVertices(
    std::array<Vector<N, Real>, (1 << N)>& vertex) const
{
    unsigned int const dsup = static_cast<unsigned int>(N);
    std::array<Vector<N, Real>, N> product;
    for (unsigned int d = 0; d < dsup; ++d)
    {
        product[d] = extent[d] * axis[d];
    }

    int const imax = (1 << N);
    for (int i = 0; i < imax; ++i)
    {
        vertex[i] = center;
        for (unsigned int d = 0, mask = 0; d < dsup; ++d, mask <<= 1)
        {
            Real sign = (i & mask ? (Real)1 : (Real)-1);
            vertex[i] += sign * product[d];
        }
    }
}
//----------------------------------------------------------------------------
template <int N, typename Real>
bool OrientedBox<N, Real>::operator==(OrientedBox const& box) const
{
    return center == box.center && axis == box.axis && extent == box.extent;
}
//----------------------------------------------------------------------------
template <int N, typename Real>
bool OrientedBox<N, Real>::operator!=(OrientedBox const& box) const
{
    return !operator==(box);
}
//----------------------------------------------------------------------------
template <int N, typename Real>
bool OrientedBox<N, Real>::operator<(OrientedBox const& box) const
{
    if (center < box.center)
    {
        return true;
    }

    if (center > box.center)
    {
        return false;
    }

    if (axis < box.axis)
    {
        return true;
    }

    if (axis > box.axis)
    {
        return false;
    }

    return extent < box.extent;
}
//----------------------------------------------------------------------------
template <int N, typename Real>
bool OrientedBox<N, Real>::operator<=(OrientedBox const& box) const
{
    return operator<(box) || operator==(box);
}
//----------------------------------------------------------------------------
template <int N, typename Real>
bool OrientedBox<N, Real>::operator>(OrientedBox const& box) const
{
    return !operator<=(box);
}
//----------------------------------------------------------------------------
template <int N, typename Real>
bool OrientedBox<N, Real>::operator>=(OrientedBox const& box) const
{
    return !operator<(box);
}
//----------------------------------------------------------------------------
